/* ccgo: igsdata.cc
 *
 * IgsData class definition, keep track of game data and player data 
 * 
 * Copyright 2002 Chun-Chung Chen <cjj@u.washington.edu>
 *
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "igsdata.hh"
#include <sstream>
#include <iostream>
using namespace std;
#include <regexx.hh>

#define atms(i, j) (rx.match[i].atom[j].str())
#define atmi(i, j) atoi(rx.match[i].atom[j].str().c_str())
#define atmf(i, j) atof(rx.match[i].atom[j].str().c_str())

GameData::GameData()
{
	number = 0;
}

GameData::GameData(const string & s)
{
	regexx::Regexx rx;
	// if seems that rx.exec can only process up to 10 atoms..
	// so we have to do this in two stages...
	if (rx.exec(s, "\\[\\s*(\\d+)\\]\\s+(\\S+) \\[\\s*(\\S+)\\s*\\] vs."
		    "\\s+(\\S+) \\[\\s*(\\S+)\\s*\\] \\((.*)\\) "
		    "\\(\\s*(\\d+)\\)") == 0)
		throw string("malformed game data string");
	number = atmi(0, 0);
	white = atms(0, 1);
	w_rank = atms(0, 2);
	black = atms(0, 3);
	b_rank = atms(0, 4);
	string nn = atms(0, 5);
	observers = atmi(0, 6);
	// this is the second stage
	if (rx.exec(nn, "(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\S+)"
		    "\\s+(\\d+) (..)") == 0) {
		throw string("malformed game data string");
	}
	moves = atmi(0, 0);
	size = atmi(0, 1);
	handicap = atmi(0, 2);
	komi = atmf(0, 3);
	byotime = atmi(0, 4);
	string ss = atms(0, 5);
	flag = (ss[0] == ' ' ? F_NORMAL :
		ss[0] == 'F' ? F_FREE :
		ss[0] == 'T' ? F_TEACH :
		ss[0] == '*' ? F_TOURNAMENT :
		F_UNKNOWN);
	type = (ss[1] == 'I' ? R_IGS_GO :
		ss[1] == 'C' ? R_CHINESE_CHESS :
		ss[1] == 'G' ? R_GOE_GO :
		ss[1] == 'P' ? R_GOE_PRO_GO :
		ss[1] == 'S' ? R_SHOGI :
		R_UNKNOWN);
}

const char * player_columns[] = {"player", "rank", "S", "idle", "Pl.", "Ob."};
const char * game_columns[] = {"No.", "White", "rank", "Black", "rank"};

IgsData::IgsData() :
	stat_players("players"),
	stat_games("games"),
	player_list(6, player_columns),
	game_list(5, game_columns),
	myid(""),
	match(0)
{
 	Gtk::ScrolledWindow * s;
	Gtk::Button * b;
	Gtk::HBox * h;
	player_list.select_row.connect
		(slot(this, & IgsData::player_list_select_row));
 	player_list.click_column.connect
 		(slot(this, & IgsData::player_list_select_col));
 	s = manage(new Gtk::ScrolledWindow());
 	s->set_policy(GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);	
 	s->add(player_list);
	player_list.set_column_justification(1, GTK_JUSTIFY_CENTER);
	player_list.set_column_justification(2, GTK_JUSTIFY_CENTER);
	player_list.set_column_justification(3, GTK_JUSTIFY_RIGHT);
	player_list.set_column_justification(4, GTK_JUSTIFY_RIGHT);
	player_list.set_column_justification(5, GTK_JUSTIFY_RIGHT);
	player_list.show();
	
	b = manage(new Gtk::Button("*"));
	h = manage(new Gtk::HBox);
	h->pack_start(* b, false);
	b->clicked.connect(player_list_refresh.slot());
	b->show();
	h->show();
	h->pack_end(stat_players);
 	pages().push_back(Gtk::Notebook_Helpers::TabElem(* s, * h));
 	s->show();

 	stat_players.show();
	
	game_list.select_row.connect
		(slot(this, & IgsData::game_list_select_row));
	game_list.click_column.connect
		(slot(this, & IgsData::game_list_select_col));
	s = manage(new Gtk::ScrolledWindow());
	s->set_policy(GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);	
	s->add(game_list);
	game_list.set_column_justification(0, GTK_JUSTIFY_RIGHT);
	game_list.set_column_justification(2, GTK_JUSTIFY_CENTER);
	game_list.set_column_justification(4, GTK_JUSTIFY_CENTER);
	game_list.show();
	b = manage(new Gtk::Button("*"));
	h = manage(new Gtk::HBox);
	h->pack_start(* b, false);
	b->clicked.connect(game_list_refresh.slot());
	b->show();
	h->show();
	h->pack_end(stat_games);
	pages().push_back(Gtk::Notebook_Helpers::TabElem(* s, * h));
	stat_games.show();
	s->show();
}

IgsData::~IgsData()
{
	for (map<int, GameData>::iterator i = games.begin(); 
	     i != games.end(); ++ i)
		if (i->second.win != NULL) {
			i->second.win->set_status("disconnected");
			i->second.win->set_end();
		}
}

void IgsData::game_update(int n, const string & w,
			  const string & b, int z)
{
	// cerr << "updating game info " << n << endl;
	GameData d;
	if (games.count(n) != 0) d = games[n];
	else {
		d.win = NULL;
		d.state = GameData::OK;
	}
	d.white = w;
	d.black = b;
	d.size = z;
	games[n] = d;
	if (d.state == GameData::NEED_WIN) {
		ready_gowin(n);
		move_skips(n);
	}
	statistics_update();
	// update game_list
	vector<string> s;
	ostringstream t;
	if (n < 100) t << ' ';
	if (n < 10) t << ' ';
	t << n;
	s.push_back(t.str());
	s.push_back(d.white);
	s.push_back(players[d.white].rank);
	s.push_back(d.black);
	s.push_back(players[d.black].rank);
	Gtk::CList_Helpers::RowList::iterator
		i = game_list.rows().find_data((void *) n); 
	double rr = game_list.get_vadjustment()->get_value();
	game_list.freeze();
	game_list.rows().insert(i, s);
	i --;
	i->set_data((void *) n);
	i ++;
	if (i != game_list.rows().end()) game_list.rows().erase(i);
	game_list.thaw();
	if (rr != game_list.get_vadjustment()->get_value())
		game_list.get_vadjustment()->set_value(rr);
}

void IgsData::game_update(int n, const GameData & gd)
{
	GameData d = gd;
	if (games.count(n) != 0) {
		GameData & dd = games[n];
		d.win = dd.win;
		d.state = dd.state;
	} else {
		d.win = NULL;
		d.state = GameData::OK;
	}
	games[n] = d;
	if (d.state == GameData::NEED_WIN) {
		ready_gowin(n);
		move_skips(n);
	}
	statistics_update();
	// update game_list
	vector<string> s;
	ostringstream t;
	if (n < 100) t << ' ';
	if (n < 10) t << ' ';
	t << n;
	s.push_back(t.str());
	s.push_back(d.white);
	s.push_back(players[d.white].rank);
	s.push_back(d.black);
	s.push_back(players[d.black].rank);
	Gtk::CList_Helpers::RowList::iterator
		i = game_list.rows().find_data((void *) n); 
	double rr = game_list.get_vadjustment()->get_value();
	game_list.freeze();
	game_list.rows().insert(i, s);
	i --;
	i->set_data((void *) n);
	i ++;
	if (i != game_list.rows().end()) game_list.rows().erase(i);
	game_list.thaw();
	if (rr != game_list.get_vadjustment()->get_value())
		game_list.get_vadjustment()->set_value(rr);
}

void IgsData::game_adjourn(int n)
{
	if (games.count(n) == 0) return;
	GameData d = games[n];
	games.erase(n);
	if (d.win != NULL) {
		d.win->set_status("game adjourned");
		d.win->set_end();
	}
	if (n == match) match = 0;
	player_update(d.white);
	player_update(d.black);
	statistics_update();
	Gtk::CList_Helpers::RowList::iterator
		i = game_list.rows().find_data((void *) n); 
	if (i != game_list.rows().end()) {
		double rr = game_list.get_vadjustment()->get_value();
		game_list.rows().erase(i);
		if (rr != game_list.get_vadjustment()->get_value())
			game_list.get_vadjustment()->set_value(rr);
	}
}

void IgsData::game_adjourn()
{
	if (match == 0) return;
	game_adjourn(match);
}

void IgsData::game_finish(int n, const string & s)
{
	if (games.count(n) == 0) return;
	GameData d = games[n];
	games.erase(n);
	if (d.win != NULL) {
		d.win->set_status(string("game finished: ") + s);
		d.win->set_end();
	}
	if (n == match) match = 0;
	player_update(d.white);
	player_update(d.black);
	statistics_update();
	Gtk::CList_Helpers::RowList::iterator
		i = game_list.rows().find_data((void *) n); 
	if (i != game_list.rows().end()) {
		double rr = game_list.get_vadjustment()->get_value();
		// game_list.freeze();
		game_list.rows().erase(i);
		// game_list.thaw();
		if (rr != game_list.get_vadjustment()->get_value())
			game_list.get_vadjustment()->set_value(rr);
	}
}

void IgsData::player_update(const string & n, const string & r,
			    PlayerData::PlayerStates s, int pl, int ob,
			    const string & id)
{
	PlayerData d;
	d.state = s;
	d.rank = r;
	d.play = pl;
	d.observe = ob;
	d.idle = id;
	players[n] = d;
	statistics_update();
	pair<const string, PlayerData> * p = & (* players.find(n));
	vector<string> ss;
	ss.push_back(n);
	ss.push_back(r);
	ss.push_back((s == PlayerData::PS_OPEN) ? " " :
		     (s == PlayerData::PS_LOOK) ? "!" :
		     (s == PlayerData::PS_CLOSE) ? "X" : "?");
	ss.push_back(id);
	ostringstream t;
	if (pl < 100) t << ' ';
	if (pl < 10) t << ' ';
	if (pl > 0) t << pl;
	ss.push_back(t.str());

	t.str("");
	if (ob < 100) t << ' ';
	if (ob < 10) t << ' ';
	if (ob > 0) t << ob;
	ss.push_back(t.str());

	Gtk::CList_Helpers::RowList::iterator i
		= player_list.rows().find_data(p);
	double rr = player_list.get_vadjustment()->get_value();
	player_list.freeze();
	player_list.rows().insert(i, ss);
	i --;
	i->set_data(p);
	i ++;
	if (i != player_list.rows().end()) player_list.rows().erase(i);
	player_list.thaw();
	if (rr != player_list.get_vadjustment()->get_value())
		player_list.get_vadjustment()->set_value(rr);
}

void IgsData::player_update(const string & n, const string & r)
{
	if (players.count(n) > 0) return;
	player_update(n, r, PlayerData::PS_UNKNOWN, 0, 0, "0s");
}

void IgsData::player_update(const string & n, const string & r, int g)
{
	if (players.count(n) > 0) {
		PlayerData d = players[n];
		player_update(n, r, d.state, g, d.observe, d.idle);
	} else {
		player_update(n, r, PlayerData::PS_UNKNOWN, g, 0, "");
	}
}

void IgsData::player_update(const string & n)
{
	if (players.count(n) == 0) return;
	// cerr << "setting " << n << "'s playing game to 0" << endl;
	PlayerData d = players[n];
	player_update(n, d.rank, d.state, 0, d.observe, d.idle);
}

void IgsData::player_remove(const string & n)
{
	map<string, PlayerData>::iterator x = players.find(n);
	if (x == players.end()) return;
	pair<const string, PlayerData> * p = & * x;
	players.erase(x);
	statistics_update();
	Gtk::CList_Helpers::RowList::iterator i
		= player_list.rows().find_data(p);
	double rr = player_list.get_vadjustment()->get_value();
	if (i != player_list.rows().end()) player_list.rows().erase(i);
	if (rr != player_list.get_vadjustment()->get_value())
		player_list.get_vadjustment()->set_value(rr);
}

void IgsData::observe_remove(int n)
{
	if (games.count(n) == 0) return;
	GameData & d = games[n];
	if (d.win != NULL) {
		d.win->set_status("removed from observation");
		d.win->set_end();
		d.win = NULL;
	}
	d.state = GameData::OK;
}

void IgsData::set_time(int n, int wt, int wb, int bt, int bb)
{
	if (ready_gowin(n)) games[n].win->set_time(wt, wb, bt, bb);
}

void IgsData::set_title(int n, const string & s)
{
	if (games.count(n) > 0) games[n].win->set_status(s);
}

void IgsData::add_move(int n, int mn, const GoGame::Move & m)
{
	if (ready_gowin(n)) {
		GameData & d = games[n];
		if (d.win->n_moves() > mn) return;
		if (d.win->n_moves() < mn) {
			move_skips.emit(n);
			return;
		}
		d.win->add_move(m);
	}
}

void IgsData::undo_move(int n)
{
	// cerr << "undoing move in " << n << endl;
	if (ready_gowin(n)) games[n].win->undo_move();
}

void IgsData::undo_move()
{
	if (match == 0) return;
	undo_move(match);
}

void IgsData::add_talk(int n, const string & w)
{
	if (ready_gowin(n)) games[n].win->kibitz(w);
}

void IgsData::add_talk(const string & w)
{
	if (match == 0) return;
	add_talk(match, w);
}

void IgsData::score_match()
{
	if (match == 0) return;
	match_win->set_status("scoring, remove dead stones");
	match_win->enter_score();
}

void IgsData::score_restore()
{
	if (match == 0) return;
	match_win->restore_board();
}

void IgsData::score_remove(const Place & p)
{
	if (match == 0) return;
	match_win->remove_group(p);
}

const string & IgsData::find_rank(const string & n)
{
	static string default_rank = "??";
	if (players.count(n) == 0) return default_rank;
	return players[n].rank;
}

void IgsData::win_close(GoWinObserve * w)
{
	for (map<int, GameData>::iterator i = games.begin(); i != games.end();
	     i ++) {
		if (i->second.win == w) {
			i->second.win = NULL;
			if (w == match_win) {
				match_win = NULL;
				match = 0;
			} else {
				i->second.state = GameData::CLOSE;
				observe_close.emit(i->first);
			}
			return;
		}
	}
}

void IgsData::set_myid(const string & s)
{
	myid = s;
}

void IgsData::realize_impl()
{
	Gtk::Notebook::realize_impl();
	adj_sizes();
}

void IgsData::style_changed_impl(Gtk::Style * s)
{
	Gtk::Notebook::style_changed_impl(s);
	adj_sizes();
}

bool IgsData::ready_gowin(int n)
{
	if (games.count(n) == 0) {
		GameData g;
		g.size = 0;
		g.state = GameData::NEED_WIN;
		g.win = NULL;
		games[n] = g;
		game_query.emit(n);
		return false;
	}
	GameData & d = games[n];
	if (d.state == GameData::CLOSE) return false;
	if (d.win == NULL) {
		if (d.size == 0) {
			if (d.state == GameData::NEED_WIN) return false;
			d.state = GameData::NEED_WIN;
			game_query.emit(n);
			return false;
		}
		bool play_w = (d.white == myid);
		bool play_b = (d.black == myid);
		if (play_w || play_b) {
			match_win = manage(new GoWinMatch(d.size,
							  play_w,
							  play_b));
			d.win = match_win;
			match = n;
			match_win->board_hit.connect(match_put.slot());
			match_win->undo_hit.connect(match_undo.slot());
			match_win->pass_hit.connect(match_pass.slot());
			match_win->done_hit.connect(match_done.slot());
			match_win->talk.connect(match_talk.slot());
		} else {
			d.win = manage(new GoWinObserve(d.size));
			d.win->talk.connect
				(SigC::bind<int>(game_talk.slot(), n));
		}
		d.win->destroy.connect
			(SigC::bind<GoWinObserve *>
			 (slot(this, & IgsData::win_close), d.win));
		ostringstream t;
		t << n;
		d.win->set_id(t.str());
		d.win->set_white(d.white);
		d.win->set_black(d.black);
		d.win->set_komi(d.komi);
		d.win->set_byotime(d.byotime);
		t.str("");
		t << "Game " << n << ": " << d.white << " [";
		if (players.count(d.white) != 0) t << players[d.white].rank;
		else t << "??";
		t << "] vs " << d.black << " [";
		if (players.count(d.black) != 0) t << players[d.black].rank;
		else t << "??";
		t << "]";
		d.win->set_title(t.str());
		d.win->set_status(d.white + " vs " + d.black);
	}
	return true;
}

void IgsData::statistics_update()
{
	ostringstream t;
	t << "Games: " << games.size();
	stat_games.set_text(t.str());
	t.str("");
	t << "Players: " << players.size();
	stat_players.set_text(t.str());
}

void IgsData::game_list_select_row(int r, int c, GdkEvent * e)
{
	if (e->type != GDK_2BUTTON_PRESS) return;
	int n = (int) game_list.row(r).get_data();
	if (games[n].win == NULL) choose_game.emit(n);
	game_list.row(r).unselect();
}

void IgsData::player_list_select_row(int r, int c, GdkEvent * e)
{
	if (e->type != GDK_2BUTTON_PRESS) return;
	string n = ((pair<const string, PlayerData> *)
		    player_list.row(r).get_data())->first;
	choose_player.emit(n);
	player_list.row(r).unselect();
}

gint compare_rank(GtkCList * cl, gconstpointer p1, gconstpointer p2)
{
	const GtkCListRow *r1 = (const GtkCListRow *) p1;
	const GtkCListRow *r2 = (const GtkCListRow *) p2;
	char * c1 = GTK_CELL_TEXT (r1->cell[cl->sort_column])->text;
	char * c2 = GTK_CELL_TEXT (r2->cell[cl->sort_column])->text;
	int i1 = atoi(c1);
	int i2 = atoi(c2);
	while (isdigit(* c1)) c1 ++;
	while (isdigit(* c2)) c2 ++;
	if (* c1 == 'p') i1 = i1 + 100;
	else if (* c1 == 'd') i1 = i1 + 50;
	else if (* c1 == 'k') i1 = 50 - i1;
	if (* c2 == 'p') i2 = i2 + 100;
	else if (* c2 == 'd') i2 = i2 + 50;
	else if (* c2 == 'k') i2 = 50 - i2;
	return ((i1 < i2) ? 1 : (i1 > i2) ? -1 : 0);
}

gint compare_num(GtkCList * cl, gconstpointer p1, gconstpointer p2)
{
	const GtkCListRow *r1 = (const GtkCListRow *) p1;
	const GtkCListRow *r2 = (const GtkCListRow *) p2;
	char * t1 = GTK_CELL_TEXT (r1->cell[cl->sort_column])->text;
	char * t2 = GTK_CELL_TEXT (r2->cell[cl->sort_column])->text;
	int i1 = atoi(t1);
	if (i1 == 0) i1 = 1000;
	int i2 = atoi(t2);
	if (i2 == 0) i2 = 1000;
	return ((i1 < i2) ? -1 : (i1 > i2) ? 1 : 0);
}

gint compare_idle(GtkCList * cl, gconstpointer p1, gconstpointer p2)
{
	const GtkCListRow *r1 = (const GtkCListRow *) p1;
	const GtkCListRow *r2 = (const GtkCListRow *) p2;
	char * c1 = GTK_CELL_TEXT (r1->cell[cl->sort_column])->text;
	char * c2 = GTK_CELL_TEXT (r2->cell[cl->sort_column])->text;
	int i1 = atoi(c1);
	int i2 = atoi(c2);
	while (isdigit(* c1)) c1 ++;
	while (isdigit(* c2)) c2 ++;
	if (* c1 == 'm') i1 *= 60;
	else if (* c1 != 's') i1 = 60 * 60;
	if (* c2 == 'm') i2 *= 60;
	else if (* c2 != 's') i2 = 60 * 60;
	return ((i1 < i2) ? -1 : (i1 > i2) ? 1 : 0);
	return 0;
}

gint compare_state(GtkCList * cl, gconstpointer p1, gconstpointer p2)
{
	const GtkCListRow *r1 = (const GtkCListRow *) p1;
	const GtkCListRow *r2 = (const GtkCListRow *) p2;
	char c1 = * GTK_CELL_TEXT (r1->cell[cl->sort_column])->text;
	char c2 = * GTK_CELL_TEXT (r2->cell[cl->sort_column])->text;
	int i1 = ((c1 == '!') ? 1 :
		  (c1 == ' ') ? 2 :
		  (c1 == 'X') ? 3 : 4);
	int i2 = ((c2 == '!') ? 1 :
		  (c2 == ' ') ? 2 :
		  (c2 == 'X') ? 3 : 4);
	return ((i1 < i2) ? -1 : (i1 > i2) ? 1 : 0);
}

void IgsData::game_list_select_col(int c)
{
	game_list.set_sort_column(c);
	if (c == 0) game_list.set_compare_func(& compare_num);
	else if (c == 2 || c == 4) game_list.set_compare_func(& compare_rank);
	else game_list.set_compare_func(NULL);
	game_list.sort();
}

void IgsData::player_list_select_col(int c)
{
	player_list.set_sort_column(c);
	switch (c) {
	case 1: player_list.set_compare_func(& compare_rank); break;
	case 2: player_list.set_compare_func(& compare_state); break;
	case 3: player_list.set_compare_func(& compare_idle); break;
	case 4:
	case 5: player_list.set_compare_func(& compare_num); break;
	default: player_list.set_compare_func(NULL);
	}
	player_list.sort();
}

void IgsData::adj_sizes()
{
	int wg = player_list.get_style()->get_font().string_width("444");
	int wr = player_list.get_style()->get_font().string_width("44k*");
	int wx = player_list.get_style()->get_font().string_width("X");
	int wn = 10 * player_list.get_style()->get_font().string_width("a");
	int wt = player_list.get_style()->get_font().string_width("44m");
	player_list.set_column_width(0, wn);
	player_list.set_column_max_width(1, wr);
	player_list.set_column_min_width(1, wr);
	player_list.set_column_max_width(2, wx);
	player_list.set_column_min_width(2, wx);
	player_list.set_column_max_width(3, wt);
	player_list.set_column_min_width(3, wt);	
	player_list.set_column_max_width(4, wg);
	player_list.set_column_min_width(4, wg);
	player_list.set_column_max_width(5, wg);
	player_list.set_column_min_width(5, wg);
	game_list.set_column_max_width(0, wg);
	game_list.set_column_min_width(0, wg);
	game_list.set_column_max_width(2, wr);
	game_list.set_column_min_width(2, wr);
	game_list.set_column_max_width(4, wr);
	game_list.set_column_min_width(4, wr);
	game_list.set_column_width(1, wn);
	game_list.set_column_width(3, wn);
}
